/*
** 		Sambar Server JAVA Engine CScript Implementation
**
**		Confidential Property of Tod Sambar
**		(c) Copyright Tod Sambar 2001
**		All rights reserved.
*/
package com.sambar.javaeng;

import java.io.*;
import java.net.*;
import java.util.*;
import java.beans.*;
import java.lang.reflect.*;

/* 
** Access to Sambar CScript API via native methods
*/
public class SambarCScript 
{
	// -------------------- Internal codes --------------------
	public final static int FAIL = 0;
	public final static int SUCCEED = 1;

	// -------------------- Log levels --------------------
	public final static int LOG_FATAL = 10;
	public final static int LOG_ERROR = 20;
	public final static int LOG_WARN = 30;
	public final static int LOG_INFO = 40;
	public final static int LOG_TRACE = 50;

	// -------------------- Result methods --------------------
	private static native void   	logmsg(long req, int level, String s);
	private static native void   	returnNULL(long req);
	private static native void   	returnString(long req, String val);
	private static native void   	returnLong(long req, long val);
	private static native void   	returnNumber(long req, double val);
	private static native void   	returnBoolean(long req, boolean val);
	private static native void   	returnObject(long req, Object val);
	private static native void   	returnException(long req, String val);

	// -------------------- CScript methods --------------------
	public static Object makeLong(long l)    
	{ 
		return new Long(l); 
	}

	public static Object makeDouble(double d)  
	{ 
		return new Double(d); 
	}

	// -------------------- Helper methods --------------------
	public static void returnResult(long req, Object val) 
	{
		if (val == null) 
		{
      		returnNULL(req);
		}
		else if (val instanceof java.lang.String) 
		{
			returnString(req, (String)val);
		} 
		else if (val instanceof java.lang.Number) 
		{
			if (val instanceof java.lang.Integer ||
				val instanceof java.lang.Short ||
				val instanceof java.lang.Byte) 
			{
				returnLong(req, ((Number)val).longValue());
			} 
			else 
			{
				/* Float, Double etc. 									*/
				returnNumber(req, ((Number)val).doubleValue());
       		}
    	} 
		else if (val instanceof java.lang.Boolean) 
		{
			returnBoolean(req, ((Boolean)val).booleanValue());
		} 
		else 
		{
			returnObject(req, val);
    	}
	}

	private static Object findMethod(Vector methods, Object args[]) 
	{
		// Only a single matching method? 
		if (methods.size() == 1) 
			return methods.firstElement();

		int maxmatch = 0;
		Object val = null;

		for (Enumeration e = methods.elements(); e.hasMoreElements(); ) 
		{
			int match = 0;
			Object element = e.nextElement();

			Class params[] = (element instanceof Method) ?
				((Method)element).getParameterTypes() :
				((Constructor)element).getParameterTypes();

			for (int i = 0; i < params.length; i++) 
			{
				Class c = params[i];
				if (c.isInstance(args[i])) 
				{
					match++;
				} 
				else if (c.isAssignableFrom(java.lang.String.class)) 
				{
					if ((args[i] instanceof byte[]) || 
						(args[i] instanceof String))
					{
						match++;
					} 
				}
        		else if (c.isPrimitive()) 
				{
					if (args[i] instanceof Number) 
					{
						if (c == Boolean.TYPE) 
							match++;
						if (c == Character.TYPE) 
							match++;
						if (c == Byte.TYPE) 
							match++;
						if (c == Short.TYPE) 
							match++;
						if (c == Integer.TYPE) 
							match++;
						if (c == Float.TYPE) 
							match++;
					} 
					else if (args[i] instanceof Boolean) 
					{
						if (c == Boolean.TYPE) 
							match++;
					} 
					else if (args[i] instanceof String) 
					{
						if (c == Character.TYPE)
							match++;
					} 
					else 
					{
						match++;
					}
				}
				else 
				{
					match++;
				}
			} 

			if (match >= maxmatch) 
			{
				maxmatch = match;
				val = element;
			}
		}

		return val;
	}

	//
	// Coerce arguments when possible to conform to the argument list.
	// Java's reflection will automatically do widening conversions,
	// unfortunately PHP only supports wide formats, so to be practical
	// some (possibly lossy) conversions are required.
	//
	private static Object[] buildArgs(Class params[], Object args[]) 
	{
		Object res[] = args;
		if (args == null)
			return res;

		for (int i = 0; i < args.length; i++) 
		{
        	Class c = params[i];

			if (args[i] instanceof byte[])
			{
				String s = new String((byte[])args[i]);
				res[i] = s;
				try 
				{
					if (c == Boolean.TYPE) 
						res[i] = new Boolean(s);
					if (c == Byte.TYPE)    
						res[i] = new Byte(s);
					if (c == Short.TYPE)   
						res[i] = new Short(s);
					if (c == Integer.TYPE) 
						res[i] = new Integer(s);
					if (c == Float.TYPE)   
						res[i] = new Float(s);
					if (c == Long.TYPE)    
						res[i] = new Long(s);
					if (c == Character.TYPE && s.length()>0) 
						res[i] = new Character(s.charAt(0));
				} 
				catch (NumberFormatException n) 
				{
				}
			} 
			else if (args[i] instanceof Number && params[i].isPrimitive()) 
			{
				if (res == args) 
					res = (Object[])res.clone();

				Number n = (Number)args[i];
				if (c == Boolean.TYPE) 
					res[i] = new Boolean(0.0 != n.floatValue());
				if (c == Byte.TYPE)    
					res[i] = new Byte(n.byteValue());
				if (c == Short.TYPE)   
					res[i] = new Short(n.shortValue());
				if (c == Integer.TYPE) 
					res[i] = new Integer(n.intValue());
				if (c == Float.TYPE)   
					res[i] = new Float(n.floatValue());
				if (c == Long.TYPE && !(n instanceof Long)) 
					res[i] = new Long(n.longValue());
			} 
		}

		return res;
	}

	// -------------------- Class creation --------------------
	public static Object create(String name, Object args[], long req) 
	{
		try 
		{
			int argslen = 0;
			if (args != null)
				argslen = args.length;

			Vector methods = new Vector();

			Constructor cons[] = Class.forName(name).getConstructors();
			if (cons.length == 0)
			{
				// No constructor for class? 
				// Return the class (i.e. java.lang.System)
				Class c = Class.forName(name);
				if (c == null)
				{
					logmsg(req, LOG_ERROR, 
						"No Java class found for " + name);
					return null;
				}

				return ((Object)c);
			}

			for (int i = 0; i < cons.length; i++) 
			{
				// Method has the same number of arguments?
				if (cons[i].getParameterTypes().length == argslen) 
					methods.addElement(cons[i]);
			}
	
			Constructor val = (Constructor)findMethod(methods, args);
			if (val == null) 
			{
				throw new InstantiationException("No Java class constructor found for " + name);
			}

			Object methargs[] = buildArgs(val.getParameterTypes(), args);
			return ((Object)val.newInstance(methargs));
		} 
		catch (Exception e) 
		{
			logmsg(req, LOG_ERROR, 
				name + ": Java class instantiation failed -- " + e.toString());
			returnException(req, e.toString());
			return null;
		}
	}

	// -------------------- Get Property --------------------
	public static int getProp(Object object, String property, long req)
	{
		try 
		{
			for (Class jclass = object.getClass(); ; jclass=(Class)object) 
			{
				// If interface is not a public class, use the first public
				// instance.
				while (!Modifier.isPublic(jclass.getModifiers())) 
				{
					Class interfaces[] = jclass.getInterfaces();
					jclass = jclass.getSuperclass();
					for (int i = interfaces.length; i-- > 0;) 
					{
						if (Modifier.isPublic(interfaces[i].getModifiers())) 
							jclass = interfaces[i];
					}
				}

				BeanInfo beanInfo = Introspector.getBeanInfo(jclass);
				PropertyDescriptor props[] = beanInfo.getPropertyDescriptors();
				for (int i = 0; i < props.length; i++)
				{
					if (props[i].getName().equalsIgnoreCase(property))
					{
						Method method = props[i].getReadMethod();
						returnResult(req, method.invoke(object, null));
						return SUCCEED;
					}
				}

				Field jfields[] = jclass.getFields();
				for (int i = 0; i < jfields.length; i++)
				{
					if (jfields[i].getName().equalsIgnoreCase(property))
					{
						returnResult(req, jfields[i].get(object));
						return SUCCEED;
					}
				}

				if (!(object instanceof Class) || (jclass == object)) 
					break;
			}

			returnException(req, property + ": Java class property not found.");
		}
		catch (Exception e) 
		{
			logmsg(req, LOG_ERROR, 
				property + ": Java class property not found -- " + 
				e.toString());
			returnException(req, e.toString());
		}

		return FAIL;
	}

	// -------------------- Invoke method --------------------
	public static int invoke(Object object, String method, 
		Object args[], long req)
	{
		try 
		{
			int argslen = 0;
			if (args != null)
				argslen = args.length;

			Vector methlist = new Vector();

			for (Class jclass = object.getClass(); ; jclass=(Class)object) 
			{
				while (!Modifier.isPublic(jclass.getModifiers())) 
				{
          			// Instance of a non-public class, substitute the first
					// public instance found.
					Class interfaces[] = jclass.getInterfaces();
					jclass = jclass.getSuperclass();
					for (int i = interfaces.length; i-- > 0;) 
					{
						if (Modifier.isPublic(interfaces[i].getModifiers())) 
							jclass = interfaces[i];
					}
				}

				Method methods[] = jclass.getMethods();
				for (int i = 0; i < methods.length; i++) 
				{
					if (methods[i].getName().equalsIgnoreCase(method))
					{
						if (methods[i].getParameterTypes().length == argslen)
							methlist.addElement(methods[i]);
					}
				}

				if (!(object instanceof Class) || (jclass == object)) 
					break;
			}

			Method val = (Method)findMethod(methlist, args);
			if (val == null) 
			{
				logmsg(req, LOG_ERROR, 
					"No Java method found for " + method);
				throw new NoSuchMethodException(method);
			}

			Object methargs[] = buildArgs(val.getParameterTypes(), args);
			returnResult(req, val.invoke(object, methargs));
			return SUCCEED;
		}
		catch (Exception e) 
		{
			logmsg(req, LOG_ERROR, 
				method + ": Java method execution failed -- " + e.toString());
			returnException(req, e.toString());
		}

		return FAIL;
	}
}
