/*
** SambarClassLoader
**
** Note: Based on AdaptiveClassLoader from www.apache.org
**
** Copyright (c) 1997-1999 The Apache Project.  All rights reserved.
*/

package com.sambar.javaeng;

import java.io.*;
import java.net.*;
import java.text.*;
import java.util.*;
import java.util.zip.*;

public class SambarClassLoader extends ClassLoader
{
    private Hashtable		classCache;
    private Vector			classRepository;
	private SambarContext	scontext;

    private static class ClassEntry 
	{
        File 	origin;
        Class	loadedClass;
        long 	lastModified;
    }

    public SambarClassLoader(SambarContext scontext, Vector classRepository, 
		boolean verify)
        throws IllegalArgumentException
    {
		super();
        classCache = new Hashtable();
        this.scontext = scontext;
        this.classRepository = classRepository;

        // Verify that all the classpath elements are valid.
		if (verify)
		{
        	Enumeration e = classRepository.elements();
        	while(e.hasMoreElements()) 
			{
        	    File file;
        	    String rname = (String)e.nextElement();

        	    try 
				{
        	    	// Check to see if element is a File instance.
            		file = new File(rname);
					if (file == null)
					{
            	    	scontext.log(SambarAPI.LOG_WARN, 
							"SambarClassLoader - Element " + rname + 
							" is not a valid \"File\" instance");
            	    	throw new IllegalArgumentException("Element " + rname + 
							" is not a valid \"File\" instance");
					}
        	    } 
				catch (Exception ex) 
				{
            	    scontext.log(SambarAPI.LOG_WARN, 
						"SambarClassLoader - Element " + rname + 
						" is not a valid \"File\" instance");
            	    throw new IllegalArgumentException("Element " + rname + 
						" is not a valid \"File\" instance");
            	}

            	// Check to see if we have proper access.
            	if (!file.exists()) 
				{
            	    scontext.log(SambarAPI.LOG_ERROR,
						"SambarClassLoader - Classpath element "
            	        + file.getAbsolutePath() + " doesn't exist!");
            	    throw new IllegalArgumentException("Classpath element "
            	        + file.getAbsolutePath() + " doesn't exist!");
            	}
				else if (!file.canRead()) 
				{
            	    scontext.log(SambarAPI.LOG_ERROR,
						"SambarClassLoader - Don't have read " +
						"access for file " + file.getAbsolutePath());
            	    throw new IllegalArgumentException("Don't have read " +
						"access for file " + file.getAbsolutePath());
            	}

            	// Check to see if it is a directory or zip/jar file
            	if (!(file.isDirectory() || isZipOrJarArchive(file))) 
				{
            	    scontext.log(SambarAPI.LOG_ERROR, "SambarClassLoader - " + 
						file.getAbsolutePath() + 
						" is not a directory or zip/jar file.");
            	    throw new IllegalArgumentException(file.getAbsolutePath()
            	        + " is not a directory or zip/jar file.");
            	}
        	}
		}
    }

    public SambarClassLoader reinstantiate()
        throws IllegalArgumentException
    {
		return new SambarClassLoader(scontext, classRepository, false);
	}

    private boolean isZipOrJarArchive(File file) 
	{
        boolean isArchive = true;
        ZipFile zipFile = null;

        try 
		{
            zipFile = new ZipFile(file);
        }
		catch (ZipException zipCurrupted) 
		{
            isArchive = false;
        }
		catch (IOException anyIOError) 
		{
            isArchive = false;
        }
		finally 
		{
            if (zipFile != null) 
			{
                try 
				{
                    zipFile.close();
                } 
				catch (IOException ignored) 
				{
				}
            }
        }

        return isArchive;
    }

    public synchronized boolean shouldReload(String classname) 
	{
        ClassEntry entry = (ClassEntry)classCache.get(classname);

        if (entry == null)
		{
            return false;
        } 
		else if (entry.origin == null)
		{
        	return false;
        } 
		else 
		{
            return (entry.origin.lastModified() != entry.lastModified);
        }
    }

    //------------------------------------ Implementation of Classloader

    /*
    ** Resolves the specified name to a Class. The method loadClass()
    ** is called by the virtual machine.  
    */
    protected synchronized Class loadClass(String name, boolean resolve)
        throws ClassNotFoundException
    {
        Class c = null;
        ClassEntry cacheEntry = (ClassEntry)classCache.get(name);
        if (cacheEntry != null) 
		{
        	// Return the cached value if already loaded
            c = cacheEntry.loadedClass;
            if (resolve) 
				resolveClass(c);
            return c;
        }

   		cacheEntry = new ClassEntry();

        // Attempt to load the class from the system
        try 
		{
        	// Throws if not found.
        	c = findSystemClass(name);
			if (c != null)
			{
        		// Add cache entry
        		cacheEntry.origin = null;
        		cacheEntry.loadedClass = c;
        		cacheEntry.lastModified = Long.MAX_VALUE;
        		classCache.put(name, cacheEntry);

                // Resolve it if necessary
        		if (resolve) 
					resolveClass(c);

        		return c;
			}
        } 
		catch (Exception e) 
		{
        }

        Enumeration repEnum = classRepository.elements();

        while (repEnum.hasMoreElements()) 
		{
            byte[] cdata = null;

            try 
			{
				String rname = (String)repEnum.nextElement();
            	File file = new File(rname);
				if (file != null)
				{
                	if (file.isDirectory())
                	    cdata = loadClassFromDirectory(file, name, cacheEntry);
                	else
                	    cdata = loadClassFromZipfile(file, name, cacheEntry);
				}
            } 
			catch(IOException ioe) 
			{
                // Error while reading in data, consider it as not found
                cdata = null;
            }
			catch (ClassCastException ce) 
			{
                // Error while reading in data, consider it as not found
                cdata = null;
            }
    
            if (cdata != null) 
			{
                // Define the class
                c = defineClass(name, cdata, 0, cdata.length);

        		// Add cache entry
                cacheEntry.loadedClass = c;
                cacheEntry.lastModified = cacheEntry.origin.lastModified();
				classCache.put(name, cacheEntry);
    
                // Resolve it if necessary
                if (resolve) 
					resolveClass(c);
                
                return c;
            }
        }

        throw new ClassNotFoundException(name);
    }

    /*
    ** Tries to load the class from a directory.
    */
    private byte[] loadClassFromDirectory(File dir, String name,
        ClassEntry cache) throws IOException
    {
        // Translate class name to file name
		//
		// Note: The SambarContext.getServlet strips the .class from
		// 		 the servlet name if it was passed to ensure this works.
        String classFileName = name.replace('.', File.separatorChar) + ".class";

        File classFile = new File(dir, classFileName);
		if (classFile == null)
			return null;

        if (classFile.exists()) 
		{
            cache.origin = classFile;
            InputStream in = new FileInputStream(classFile);
            try 
			{
                return loadBytesFromStream(in, (int) classFile.length());
            } 
			finally 
			{
                in.close();
            }
        }
		else 
		{
            // Not found
            return null;
        }
    }

    /*
    ** Tries to load the class from a zip file.
    */
    private byte[] loadClassFromZipfile(File file, String name,
    	ClassEntry cache) throws IOException
    {
        // Translate class name to file name
        String classFileName = name.replace('.', File.separatorChar) + ".class";

        ZipFile zipfile = new ZipFile(file);

        try 
		{
            ZipEntry entry = zipfile.getEntry(classFileName);
            if (entry != null) 
			{
                cache.origin = file;
                return loadBytesFromStream(zipfile.getInputStream(entry),
                    (int) entry.getSize());
            } 
			else 
			{
                // Not found
                return null;
            }

        }
		finally 
		{
            zipfile.close();
        }
    }

    /**
     * Loads all the bytes of an InputStream.
     */
    private byte[] loadBytesFromStream(InputStream in, int length)
        throws IOException
    {
        byte[] buf = new byte[length];
        int nRead, count = 0;

        while ((length > 0) && ((nRead = in.read(buf,count,length)) != -1)) {
            count += nRead;
            length -= nRead;
        }

        return buf;
    }

    /*
    ** Get an InputStream on a given resource.  Will return null if no
    ** resource with this name is found.
    */
    public InputStream getResourceAsStream(String name) 
	{
        // Try to load it from the system class
        InputStream s = getSystemResourceAsStream(name);

        if (s == null) 
		{
            // Try to find it from every repository
            Enumeration repEnum = classRepository.elements();
            while (repEnum.hasMoreElements()) 
			{
                File file = (File) repEnum.nextElement();
                if (file.isDirectory())
                    s = loadResourceFromDirectory(file, name);
                else
                    s = loadResourceFromZipfile(file, name);

                if (s != null)
                    break;
            }
        }

        return s;
    }

    /*
    ** Loads resource from a directory.
    */
    private InputStream loadResourceFromDirectory(File dir, String name) 
	{
        // Name of resources are always separated by /
        String fileName = name.replace('.', File.separatorChar);
        File resFile = new File(dir, fileName);

        if (resFile.exists()) 
		{
            try
			{
                return new FileInputStream(resFile);
            } 
			catch (FileNotFoundException shouldnothappen) 
			{
                return null;
            }
        } 
		else 
		{
            return null;
        }
    }

    /*
    ** Loads resource from a zip file
    */
    private InputStream loadResourceFromZipfile(File file, String name) 
	{
        try 
		{
            ZipFile zipfile = new ZipFile(file);
            ZipEntry entry = zipfile.getEntry(name);

            if (entry != null)
                return zipfile.getInputStream(entry);
            else
                return null;
        }
		catch(IOException e) 
		{
            return null;
        }
    }

    /*
    ** Find a resource with a given name.  The return is a URL to the
    ** resource. Doing a getContent() on the URL may return an Image,
    ** an AudioClip,or an InputStream.
	*/
    public URL getResource(String name) 
	{
        URL u = getSystemResource(name);
        if (u != null) 
            return u;

        Enumeration repEnum = classRepository.elements();
        while (repEnum.hasMoreElements()) 
		{
            File file = (File) repEnum.nextElement();
            if (file.isDirectory()) 
			{
                String fileName = name.replace('.', File.separatorChar);
                File resFile = new File(file, fileName);
                if (resFile.exists()) 
				{
                    // Build a file:// URL form the file name
                    try 
					{
                        return new URL("file://" + resFile.getAbsolutePath());
                    } 
					catch(java.net.MalformedURLException badurl) 
					{
                        badurl.printStackTrace();
                        return null;
                    }
                }
            }
        }

        // Not found
        return null;
    }
}
