/*
**      Sambar Server multipart/form-data request processor
**
**	Sample use:
**
**		MultipartRequest mreq = new MultipartRequest(req, "c:/temp");
**
**		String p1 = mreq.getParameter("p1");
**		String p2 = mreq.getParameter("p2");
**		MultipartFile f = mreq.getFile("uploadfile");
**		:::
**		:::
**		// Delete the temporary file(s)
**		mreq.deleteFiles();	
**
**	Note:
**		This package was modeled after Jason Hunter's MultipartRequest
**		handler (www.servlets.com).
*/
package com.sambar.multipart;

import java.io.*;
import java.util.*;
import javax.servlet.*;
import javax.servlet.http.*;

public class MultipartRequest 
{
	private static final int MAX_UPLOAD_SIZE = 1024 * 1024;

	private File dir;
	private Hashtable parameters = new Hashtable();  
	private Hashtable files = new Hashtable();       

	public MultipartRequest(HttpServletRequest req, String dirpath) 
		throws IOException 
	{
		this(req, dirpath, MAX_UPLOAD_SIZE);
	}

	public MultipartRequest(HttpServletRequest req, String dirpath, int maxsize)
		throws IOException 
	{
		if (dirpath == null)
			throw new IllegalArgumentException("dirpath cannot be null");

		// Save the dir
		this.dir = new File(dirpath);

		// Check saveDirectory is truly a directory
		if (!dir.isDirectory())
			throw new IllegalArgumentException("Not a directory: " + dirpath);

		// Check dir is writable
		if (!dir.canWrite())
			throw new IllegalArgumentException("Not writable: " + dirpath);

		//
		// Parse the request saving data to "parameters" and "files";
		// write the file contents to the directory.
		//

		int len = req.getContentLength();
		if (len > maxsize) 
		{
			throw new IOException("Content length of " + len + 
				" exceeds " + maxsize + " bytes");
		}

		// Verify the content type is "multipart/form-data"
		String type = req.getContentType();
		if ((type == null) || 
			(!type.toLowerCase().startsWith("multipart/form-data")))
		{
			throw new IOException("Content type is not multipart/form-data");
		}

		//
		// Get the boundary string.
		//
		int index = type.lastIndexOf("boundary=");
		if (index == -1) 
			throw new IOException("Separation boundary was not specified");

		String boundary = type.substring(index + 9);
		boundary = "--" + boundary;

		MultipartStream in = new MultipartStream(req.getInputStream(), len);

		// Read the first line (first boundary)
		String line = in.readLine(true);
		if (line == null) 
			throw new IOException("Corrupt form data: premature ending");

		// Verify that the line is the boundary
		if (!line.startsWith(boundary)) 
			throw new IOException("Corrupt form data: no leading boundary");

		// Process each boundary segment
		boolean done = false;
		while (!done) 
		{
			done = processForm(in, boundary);
		}
	}

	public Enumeration getParameterNames() 
	{
		// Return the parsed parameter names
		return parameters.keys();
	}

	public Enumeration getFileNames() 
	{
   		// Returns the names of the uploaded files
		return files.keys();
	}

	public String getParameter(String name) 
	{
		try 
		{
			Vector values = (Vector)parameters.get(name);
			if (values == null || values.size() == 0) 
				return null;

			String value = (String)values.elementAt(values.size() - 1);
			return value;
		}
		catch (Exception e) 
		{
			return null;
		}
	}

	public String[] getParameterValues(String name) 
	{
		try 
		{
			Vector values = (Vector)parameters.get(name);
			if (values == null || values.size() == 0)
				return null;
      
			String[] valuesArray = new String[values.size()];
			values.copyInto(valuesArray);
			return valuesArray;
		}
		catch (Exception e) 
		{
			return null;
		}
	}

	public MultipartFile getFile(String name) 
	{
		try 
		{
			return (MultipartFile)files.get(name);
		}
		catch (Exception e) 
		{
			return null;
		}
	}

	public void deleteFiles() 
	{
		MultipartFile f;
		Enumeration allfiles;

		allfiles = files.elements();
		while (allfiles.hasMoreElements())
		{
			f = (MultipartFile)allfiles.nextElement();
			f.delete();
		}
	}

	protected boolean processForm(MultipartStream in, String boundary) 
		throws IOException 
	{
		// Read the first line (if no line is found, we're done)
		//
		// Format:
		//
		//    content-disposition: form-data; name="field"; filename="file.txt"
		//
		String line = in.readLine(true);
		if ((line == null) || (line.length() == 0))
			return true;

		//
		// Parse the content-disposition line (disposition, name, filename)
		//
		boolean nofile = false;

		//
		// Get the content disposition ("form-data")
		//
		String tmpline = line.toLowerCase();
		int start = tmpline.indexOf("content-disposition: ");
		int end = tmpline.indexOf(";");
		if ((start == -1) || (end == -1))
			throw new IOException("Content disposition corrupt: " + line);

		String disposition = tmpline.substring(start + 21, end);
		if (!disposition.equals("form-data"))
			throw new IOException("Invalid content disposition: "+ disposition);

		//
		// Get the field name
		//
		start = tmpline.indexOf("name=\"", end);
		end = tmpline.indexOf("\"", start + 7);
		if ((start == -1) || (end == -1))
			throw new IOException("Content disposition corrupt: " + line);

		String name = line.substring(start + 6, end);

		//
		// Get the file name, if specified
		//
		String filename = null;
		start = tmpline.indexOf("filename=\"", end + 2);
		end = tmpline.indexOf("\"", start + 10);
		if ((start != -1) && (end != -1))
		{
			filename = line.substring(start + 10, end);

			// The filename may contain a full path.  
			// Cut to just the filename.
			int slash = Math.max(filename.lastIndexOf('/'), 
				filename.lastIndexOf('\\'));
			if (slash > -1)
				filename = filename.substring(slash + 1);

			if (filename.equals("")) 
				nofile = true;
		}

		//
		// Get the content-type, if specified
		//
		line = in.readLine(true);
		if (line == null) 
			return true;

		String contentType = null;
		tmpline = line.toLowerCase();
		if (tmpline.startsWith("content-type")) 
		{
			start = tmpline.indexOf(" ");
			if (start == -1) 
				throw new IOException("Content type corrupt: " + line);

			contentType = line.substring(start + 1);

			// Skip the empty line
			line = in.readLine(true);
			if ((line == null) || (line.length() > 0))
			{
				throw new IOException("No blank line after content type: " + 
					line);
			}
		}
		else 
		{
			// no content type, so should be empty
			if (line.length() != 0) 
			{  
				throw new IOException("No blank line after disposition: " + 
					line);
			}

			// Assume a default content type
			contentType = "application/octet-stream";
		}

		//
		// Read the content (end after reading the boundary)
		//
		if (filename == null) 
		{
			String value = readParameter(in, boundary);
			if (value.equals("")) 
			{
				// Empty strings ==> null
				value = null;  
			}

			Vector curvalues = (Vector)parameters.get(name);
			if (curvalues == null) 
			{
				curvalues = new Vector();
				parameters.put(name, curvalues);
			}

			curvalues.addElement(value);
		}
		else 
		{
			readAndSaveFile(in, boundary, filename, contentType, nofile);
			if (!nofile)
			{
				files.put(name, new MultipartFile(dir.toString(), filename, 
					contentType));
			}
		}

		// there's more to read
		return false;
	}

	protected String readParameter(MultipartStream in, String boundary) 
		throws IOException 
	{
		StringBuffer sbuf = new StringBuffer();
		String line;

		while ((line = in.readLine(true)) != null) 
		{
			if (line.startsWith(boundary)) 
				break;
			sbuf.append(line + "\r\n");  
		}

		if (sbuf.length() == 0) 
			return null;

		// Trim the last line's \r\n
		sbuf.setLength(sbuf.length() - 2);  
		return sbuf.toString();
	}

	protected void readAndSaveFile(MultipartStream in, String boundary, 
		String filename, String contentType, boolean ignore) 
		throws IOException 
	{
		OutputStream out = null;

		// If no file name as specified/sent, read to the next boundary 
		// and ignore the empty contents
		if (ignore)
		{
			out = new ByteArrayOutputStream();
		}
		else 
		{
			File f = new File(dir + File.separator + filename);
			out = new BufferedOutputStream(new FileOutputStream(f), 8 * 1024);
		}

		in.readUntilBoundary(out, boundary);
		out.flush();
		out.close();

		// Throw away the boundary line
		String line = in.readLine(true);
	}
}

class MultipartStream 
{
	ServletInputStream bufInput;
	byte[]             bufBytes;
	int                bufPos, bufLen, bufLeft;

	public MultipartStream(ServletInputStream in, int totalExpected) 
	{
		// set input buffer
		this.bufInput = in;
		this.bufPos = 0;
		this.bufLen = 0;
		this.bufLeft = totalExpected;
		this.bufBytes = new byte[4096];
	}

    /**
     *  Read the first 255 characters of the current line on the servlet 
	 *	input stream
     *
     *  @param proceedToNextLine is true when the current line should be set
     *  to the next one
     *   
     *  @returns the line read or <TT>null<TT> if there is no more input
    */
    public String readLine(boolean proceedToNextLine) 
	throws IOException 
    {
        int     i = 0;
        String	res;

        if (bufLen - bufPos < 256) 
		{
            System.arraycopy(bufBytes, bufPos, bufBytes, 0, bufLen - bufPos);
            bufLen -= bufPos;
            bufPos = 0;
            fillBuffer();
        }

        if (bufPos == bufLen)
            return null;

        for(i = bufPos; i < bufPos + 255 && i < bufLen; i++) 
		{
            if(bufBytes[i] == '\r' && bufBytes[i+1] == '\n') 
			{
                res = new String(bufBytes, bufPos, i - bufPos);
                if(proceedToNextLine)
                    bufPos = i+2;
                return res;
            }
        }

        res = new String(bufBytes, bufPos, i - bufPos);
        if (proceedToNextLine) 
		{
            boolean justSawCR = false;

            bufPos = i;
            for (;;) 
			{
                if (bufPos >= bufLen) 
				{
                    bufPos = 0;
                    bufLen = 0;
                    if (!fillBuffer())
                        break;
                }

                if (bufBytes[bufPos] == '\r') 
				{
                    justSawCR = true;
                } 
				else if(bufBytes[bufPos] == '\n' && justSawCR) 
				{
                    bufPos++;
                    break;
                } 
				else 
				{
                    justSawCR = false;
                }
                bufPos++;
            }
        }

        return res;
    }

    public int readUntilBoundary(OutputStream out, String boundary) 
		throws IOException 
    {
        boolean	first = true;
        String	line;
        int		boundaryLen;
        int		lineLen;
        int		totalLength = 0;

        boundaryLen = boundary.length();
        for (;;) 
		{
            line = readLine(false);
            if(line == null)
                return -1;

            if (line.startsWith(boundary)) 
			{
                lineLen = line.length();
                if ((lineLen == boundaryLen) ||
                    ((lineLen == boundaryLen + 2) && 
		             (line.charAt(boundaryLen) == '-') && 
                     (line.charAt(boundaryLen+1) == '-'))) 
			    {
                    return totalLength;
                }
        	}

            if (first) 
			{
                first = false;
            } 
			else 
			{
                out.write((char) '\r');
                out.write((char) '\n');
                totalLength += 2;
            }
            totalLength += readLineFully(out);
        }
    }

    private int readLineFully(OutputStream out) 
		throws IOException 
    {
        boolean	justSawCR = false;
        int		lastBufPos = bufPos;
        int		totalLength = 0;

        for (;;) 
		{
            if (bufPos >= bufLen) 
			{
                if(lastBufPos < bufPos) 
				{
                    out.write(bufBytes, lastBufPos, bufPos - lastBufPos);
                    totalLength += bufPos - lastBufPos;
                }

                lastBufPos = 0;
                bufPos = 0;
                bufLen = 0;
                if (!fillBuffer())
                    break;
            }

            if (bufBytes[bufPos] == '\r') 
			{
                justSawCR = true;
            } 
			else if (bufBytes[bufPos] == '\n' && justSawCR) 
			{
                if (lastBufPos < bufPos-1) 
				{
                    out.write(bufBytes, lastBufPos, (bufPos-1) - lastBufPos);
                    totalLength += (bufPos-1) - lastBufPos;
                }
                bufPos++;
                break;
            } 
			else 
			{
                justSawCR = false;
            }
            bufPos++;
        }
        return totalLength;
    }

    private boolean fillBuffer() 
	throws IOException 
    {
        int		nrRead;
        boolean	readSomething = false;

        while (bufLeft > 0 && bufLen < bufBytes.length) 
		{
            nrRead = bufInput.read(bufBytes, bufLen, 
			Math.min(bufBytes.length - bufLen, bufLeft));
            if (nrRead >= 0) 
			{
                bufLen += nrRead;
                bufLeft -= nrRead;
                readSomething = true;
            }
        }

        return readSomething;
    }
}
